/* OpenVAS-Client
 *
 * Description: Utilities for priority management
 *
 * Authors:
 * Joey Schulze <joey@infodrom.org>
 *
 * Copyright:
 * Copyright (C) 2009 Intevation GmbH
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * or, at your option, any later version as published by the Free
 * Software Foundation
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * In addition, as a special exception, you have
 * permission to link the code of this program with the OpenSSL
 * library (or with modified versions of OpenSSL that use the same
 * license as OpenSSL), and distribute linked combinations including
 * the two. You must obey the GNU General Public License in all
 * respects for all of the code used other than OpenSSL. If you
 * modify this file, you may extend this exception to your version
 * of the file, but you are not obligated to do so. If you do not
 * wish to do so, delete this exception statement from your version.
 */

/** @file
 * Priority management.
 * Shows dialogs to allow the user to change the priority of NVTS, e.g. to mark
 * False  Positives or to emphasize underestemated threats (mark a certain
 * message as hole).
 */

#include <stdlib.h>
#include <string.h>
#include "context.h"
#include "parser.h"
#include "parseutils.h"
#include "priority.h"
#include "data_mining.h"

/**
 * @brief Free priority chain.
 * 
 * @param priority Priority chain to free.
 */
void
priority_free(struct priority *priority)
{
  struct priority *next = priority;

  while (priority) {
    next = priority->next;
    efree(priority);
    priority = next;
  }
}

/**
 * @brief Add priority to chain.
 *
 * Add a priority triple (port,prio old and new) to an existing
 * priority chain at the proper location inside.  A new priority pair
 * for the same port will be added in sequence which is needed for
 * later list generation.
 *
 * The priority chain must always contain at least one element and
 * thus not be NULL.
 * 
 * @param priority Priority chain.
 * @param element New priority to add.
 */
void
priority_add(struct priority *list, struct priority *element)
{
  struct priority *pivot = list;

  if (!list)
    return;

  while (pivot && strcmp(pivot->port, element->port))
    pivot = pivot->next;

  if (pivot) {
    element->next = pivot->next;
    pivot->next = element;
  } else {
    pivot = list;
    while (pivot->next)
      pivot = pivot->next;
    element->next = NULL;
    pivot->next = element;
  }
}

/**
 * @brief Determine whether a priority as been edited by the user.
 *
 * The priority chain is searched for an element with edited priority.
 * 
 * @param priority Priority chain.
 *
 * @return 1 (true) or 0 (false).
 */
int
priority_has_changed(struct priority *priority)
{
  struct priority *pivot = priority;

  while (pivot) {
    if (pivot->prio_old != pivot->prio_new)
      return 1;
    pivot = pivot->next;
  }

  return 0;
}

/**
 * @brief Count number of priority elements in chain
 *
 * @param priority Priority chain
 *
 * @return number of priorities
 */
int
priority_count(struct priority *priority)
{
  int i = 0;

  if (!priority)
    return 0;

  while (priority) {
    i++;
    priority = priority->next;
  }

  return i;
}

/**
 * @brief Determine whether the priority chain contains multiple ports
 *
 * An OpenVAS plugin script may test multiple ports and thus the
 * priority chain may contain priority pairs for more than only one
 * port.  This function determines whether the chain contains pairs
 * for different ports.  This function requires a priority chain built
 * up by priority_add().
 *
 * @param priority Priority chain
 *
 * @return 1 (true) or 0 (false)
 */
int
priority_is_multiport(struct priority *priority)
{
  char *port;

  if (!priority)
    return 0;

  port = priority->port;

  while (priority) {
    if (strcmp(port, priority->port))
      return 1;

    priority = priority->next;
  }

  return 0;
}

/**
 * @brief Find a matching priority element.
 *
 * An OpenVAS plugin script may test multiple ports and thus the
 * priority chain may contain priority pairs for more than only one
 * port.  This function tries to find the priority element matching
 * the parameters.  This function requires a priority chain built
 * up by priority_add().
 *
 * This function searches a priority element with the same priorities
 * as specified unless they are -1.
 *
 * @param priority Priority chain
 * @param port Port to match
 * @param prio_old to match or -1
 * @param new_old to match or -1
 *
 * @return matching priority element or NULL
 */
struct priority *
priority_find(struct priority *priority, char *port, int prio_old, int prio_new)
{
  while (priority && priority->port &&
	 (strcmp(priority->port, port) ||
	  (prio_old != -1 && priority->prio_old != prio_old) ||
	  (prio_new != -1 && priority->prio_new != prio_new)))
    priority = priority->next;

  return priority;
}

/**
 * @brief Parse a priority configuration record.
 *
 * This function interprets a configuration record for one OpenVAS ID
 * as found in the scope configuration.
 *
 * @param preference Configuration string.
 *
 * @return priority chain with all priority elements.
 */
struct priority *
priority_parse_preference (char *preference)
{
  char *pivot;
  char *port;
  char *confprio;
  int prio_old, prio_new;
  struct priority *priority = NULL;
  struct priority *current = NULL, *item;

  if (!preference || !strlen(preference))
    return NULL;

  pivot = preference;

  while (strlen(pivot) > 5) {
    port = parse_separator(pivot);

    if (!port)
      return priority;

    pivot += strlen(port) + 5;
    confprio = parse_separator(pivot);

    if (!confprio) {
      efree(&port);
      return priority;
    }

    if ((prio_old = priority_name_to_type(confprio)) == -1) {
      efree(&port);
      efree(&confprio);
      return priority;
    }

    pivot += strlen(confprio) + 5; 
    efree(&confprio);

    confprio = parse_separator(pivot);

    if (!confprio) {
      efree(&port);
      return priority;
    }

    if ((prio_new = priority_name_to_type(confprio)) == -1) {
      efree(&port);
      efree(&confprio);
      return priority;
    }

    pivot += strlen(confprio) + 5; 
    efree(&confprio);

    item = emalloc(sizeof(struct priority));
    item->port = port;
    item->prio_old = prio_old;
    item->prio_new = prio_new;
    item->next = NULL;

    if (priority)
      priority_add(priority, item);
    else
      priority = item;
    current = item;
  }

  return priority;
}

/**
 * @brief Adjust the priority chain from the report to reality.
 *
 * This function takes the priority chain gathered from the current
 * report as input.  It then reads the report configuration and
 * tries to calculate priorities back to the original ones.  Next the
 * scope configuration is read and edited priorities added to the
 * chain.
 *
 * @param priolist[in,out] Priority chain.
 *
 * @return The supplied priority chain is modified.
 */
void
priorities_adjust (struct arglist *priolist)
{
  struct context *report, *scope;
  struct arglist *preferences;
  struct arglist *hostlist;
  struct priority *pref_priority;
  struct priority *priority;
  struct priority *pivot;
  char prefkey[200];

  if ((report = context_by_type(Context, CONTEXT_REPORT)) == NULL)
    return;

  if ((scope = context_by_type(Context, CONTEXT_SCOPE)) == NULL)
    return;

  while (priolist->next) {
    hostlist = priolist->value;
    snprintf(prefkey, sizeof(prefkey), "PLUGINS_PRIORITIES_%s", priolist->name);

    if ((preferences = arg_get_value(report->prefs, prefkey)) != NULL) {
      while (preferences->next) {
	if ((pref_priority = priority_parse_preference((char *)preferences->value)) == NULL)
	  break;

	while (pref_priority) {
	  if ((priority = arg_get_value(hostlist, preferences->name))) {
	    if ((pivot = priority_find(priority, pref_priority->port, pref_priority->prio_new, -1)))
	      pivot->prio_old = pivot->prio_new = pref_priority->prio_old;
	  }

	  pref_priority = pref_priority->next;
	}
	priority_free(pref_priority);
	preferences = preferences->next;
      }

    }

    if ((preferences = arg_get_value(scope->prefs, prefkey)) != NULL) {
      while (preferences->next) {
	if ((pref_priority = priority_parse_preference((char *)preferences->value)) == NULL)
	  break;

	while (pref_priority) {
	  if ((priority = arg_get_value(hostlist, preferences->name))) {
	    if ((pivot = priority_find(priority, pref_priority->port, -1, pref_priority->prio_old))) {
	      pivot->prio_old = pref_priority->prio_old;
	      pivot->prio_new = pref_priority->prio_new;
	    } else {
	      pivot = emalloc(sizeof(struct priority));
	      pivot->port = estrdup(pref_priority->port);
	      pivot->prio_old = pref_priority->prio_old;
	      pivot->prio_new = pref_priority->prio_new;
	      pivot->next = NULL;
	      priority_add(priority, pivot);
	      /*
	      pivot->next = priority->next;
	      priority->next = pivot;
	      */
	    }
	  } else {
	    pivot = emalloc(sizeof(struct priority));
	    pivot->port = estrdup(pref_priority->port);
	    pivot->prio_old = pref_priority->prio_old;
	    pivot->prio_new = pref_priority->prio_new;
	    pivot->next = NULL;
	    arg_add_value(hostlist, estrdup(preferences->name), ARG_PTR, sizeof(struct priority), pivot);
	  }

	  pref_priority = pref_priority->next;
	}
	priority_free(pref_priority);
	preferences = preferences->next;
      }

    }

    priolist = priolist->next;
  }
  return;

  if ((preferences = arg_get_value(report->prefs, "PLUGINS_PRIORITIES")) == NULL)
    return;

  if ((preferences = arg_get_value(scope->prefs, "PLUGINS_PRIORITIES")) == NULL)
    return;

}

/**
 * @brief Determine all visible OIDs and their priorities.
 *
 * This function reads the oids and priorities for all hosts from the
 * current report and adjusts priorities on the base of the report and
 * scope configuration.
 *
 * @return Priority chains for all visible hosts.
 */
struct arglist *
get_priorities (struct arglist *ctrls)
{
  struct arglist *priolist;
  struct arglist *hostlist;
  struct subset *walk;
  struct priority *item;
  struct priority *pivot;
  char *oid, *port, *severity, *host;
  int be = GPOINTER_TO_SIZE(arg_get_value(ctrls, "BE"));
  int prio;

  priolist = emalloc(sizeof(struct arglist));

  walk = query_backend(be, "SELECT plugin_oid,port,severity,host FROM results");

  while (walk) {
    oid = subset_nth_value(walk, 0);
    port = subset_nth_value(walk, 1);
    severity = subset_nth_value(walk, 2);
    host = subset_nth_value(walk, 3);
    prio = priority_str_to_type(severity);

    printf ("BUGME: priorities found\n");
    printf ("BUGME: \t oid %s\n", oid);
    printf ("BUGME: \t port  %s\n", port);
    printf ("BUGME: \t severity %s\n", severity);
    printf ("BUGME: \t host %s\n", severity);
    printf ("BUGME: \t prio: %d\n", prio);
    
    if ((hostlist = arg_get_value(priolist, host)) == NULL) {
      hostlist = emalloc(sizeof(struct arglist));
      arg_add_value(priolist, host, ARG_ARGLIST, -1, hostlist);
    }

    if ((item = arg_get_value(hostlist, oid)) == NULL) {
      item = emalloc(sizeof(struct priority));
      item->port = port;
      item->prio_old = item->prio_new = prio;
      item->next = NULL;

      arg_add_value(hostlist, oid, ARG_PTR, sizeof(struct priority), item);
    } else {
      pivot = item;
      while (pivot && (strcmp(pivot->port, port) || pivot->prio_old != prio))
	pivot = pivot->next;

      if (pivot == NULL) {
	pivot = emalloc(sizeof(struct priority));
	pivot->port = port;
	pivot->prio_old = pivot->prio_new = prio;
	pivot->next = NULL;

	priority_add(item, pivot);
      }
    }

    walk = subset_next(walk);
  }

  priorities_adjust(priolist);

  return priolist;
}

/**
 * @brief Build a configuration string for an OID.
 *
 * This function generates a configuration line for all modified
 * priorities in the given chain as used in the scope configuration.
 *
 * This function should only be called if priority_has_changed()
 * returns true.
 *
 * @param priority Priority chain.
 *
 * @return String with the configuration.
 */
char *
format_priorities(struct priority *priority)
{
  struct priority *pivot = priority;
  char *str = NULL;
  char *tmp;
  char *prio_old, *prio_new;

  while (pivot) {
    if (pivot->prio_old != pivot->prio_new) {
      if (str == NULL)
	str = estrdup("<|>");

      prio_old = priority_type_to_name(pivot->prio_old);
      prio_new = priority_type_to_name(pivot->prio_new);
      tmp = emalloc(strlen(str) + strlen(pivot->port) + strlen(prio_old) +
		    strlen(prio_new) + 3*5 + 1);
      sprintf(tmp, "%s %s <|> %s <|> %s <|>", str, pivot->port, prio_old, prio_new);
      efree(&str);
      str = tmp;
    }
    pivot = pivot->next;
  }

  return str;
}

/**
 * @brief Free the priorities data structure
 *
 * This function frees the data structure built by get_priorities()
 *
 * @param priorities List to be freed
 */
void
priorities_free(struct arglist *priorities)
{
  struct arglist *pivot = priorities;
  struct arglist *v;

  if (!pivot)
    return;

  while (pivot->next) {

    v = pivot->value;
    while (v->next) {
      priority_free(v->value);
      v->value = NULL;

      v = v->next;
    }
    arg_free_all(pivot->value);
    pivot->value = NULL;
    pivot = pivot->next;
  }

  arg_free_all(priorities);
}

/**
 * @brief Delete a configuration entry.
 *
 * The line to be removed in the scope configuration is determined by
 * the hostname and the oid.  If the last configuration entry has been
 * removed the entire block is deleted from the scope preferences.
 *
 * @param host Hostname
 * @param oid OpenVAS ID
 */
void
preference_delete(char *host, char *oid)
{
  struct context *scope;
  struct arglist *preferences;
  char *conf;
  gchar *prefkey;

  if (!oid)
    return;

  if ((scope = context_by_type(Context, CONTEXT_SCOPE)) == NULL)
    return;

  prefkey = g_strdup_printf ("PLUGINS_PRIORITIES_%s", host);

  if ((preferences = arg_get_value(scope->prefs, prefkey)) == NULL) {
    g_free (prefkey);
    return;
  }

  if ((conf = arg_get_value(preferences, oid)) == NULL) {
    g_free (prefkey);
    return;
  }

  efree(&conf);

  arg_del_value(preferences, oid);

  if (preferences->next == NULL) {
    efree(&preferences->value);
    arg_del_value(scope->prefs, prefkey);
  }

  g_free (prefkey);
}

/**
 * @brief Store a (new) priority override in the scope preferences.
 *
 * A configuration line is generated for the given priority chain and
 * then stored in the scope configuration.
 *
 * @param host Hostname
 * @param oid OpenVAS ID
 * @param priority Priority chain
 */
void
preference_store (char *host, char *oid, struct priority *priority)
{
  struct context *scope;
  struct arglist *preferences;
  char *conf;
  gchar *prefkey;
  char *priorities_str;

  if (!host || !oid || !priority)
    return;

  if ((scope = context_by_type(Context, CONTEXT_SCOPE)) == NULL)
    return;

  if (!priority_has_changed(priority)) {
    preference_delete(host, oid);
    return;
  }

  prefkey = g_strdup_printf ("PLUGINS_PRIORITIES_%s", host);

  priorities_str = format_priorities(priority);

  if ((preferences = arg_get_value(scope->prefs, prefkey)) == NULL) {
    struct arglist *preferences = emalloc(sizeof(struct arglist));
    arg_add_value(preferences, oid, ARG_STRING, strlen(priorities_str)+1, priorities_str);
    arg_add_value(scope->prefs, prefkey, ARG_ARGLIST, -1, preferences);
    g_free (prefkey);
    return;
  }

  if ((conf = arg_get_value(preferences, oid)) == NULL)
    arg_add_value(preferences, oid, ARG_STRING, strlen(priorities_str)+1, priorities_str);
  else
    arg_set_value(preferences, oid, strlen(priorities_str)+1, priorities_str);

  g_free (prefkey);
}
